//: Playground - noun: a place where people can play

import Cocoa

var str = "Hello, playground"

enum CodingError: Error {               // enum for coding errors
    case decoding(String)
}

enum Happiness: Codable {               // our enum is "Codable"
    case Sad
    case Happy(ofGrade: Int)
    
    private enum CodingKeys: String, CodingKey {    // the keys for the coders are not simple strings anymore
        case Happiness                  // but enums of Strings (or Ints)
        case Grade
    }

    init(from decoder: Decoder) throws {    // our init throws errors
        let values = try decoder.container(keyedBy: CodingKeys.self)    // get the container from the decoder
        if let happiness = try? values.decode(Int.self, forKey: .Happiness) {    // this is the main value of the enum
            switch happiness {          // and look at it
            case 0:                     // its the first member of the enum
                self = .Sad             // with no assciated value
                return                  // no need to continue
            case 1:                     // the second one
                if let grade = try? values.decode(Int.self, forKey: .Grade) {   // decode the associated value
                    self = .Happy(ofGrade: grade)   // and set the enum
                    return
                }
            default:                    // should never happen
                break
            }
        }
        throw CodingError.decoding("Decoding Error: \(dump(values))")   // had no success
    }
    
    func encode(to encoder: Encoder) throws {   // the encoder thows errors as well
        var happiness: Int              // lets map the cases of the enum to an Int
        var container = encoder.container(keyedBy: CodingKeys.self) // get an empty container first
        switch self {                   // and look at our enum
        case .Sad:                      // map first case to 0
            happiness = 0               // no associated value to encode
        case .Happy(let grade):         // map second case to 1
            happiness = 1
            try container.encode(grade, forKey: .Grade) // encode the associated value
        }
        try container.encode(happiness, forKey: .Happiness) // and encode the main value of the enum
    }
}

let mood = Happiness.Happy(ofGrade: 42)
let data = try PropertyListEncoder().encode(mood)
let mood2 = try PropertyListDecoder().decode(Happiness.self, from: data)
Swift.print("\(mood2)")

let data1 = NSKeyedArchiver.archivedData(withRootObject: data)
if let data11 = NSKeyedUnarchiver.unarchiveObject(with: data1) as? Data,
    let mood1 = try? PropertyListDecoder().decode(Happiness.self, from: data11) {
    Swift.print("\(mood1)")
}

struct Person: Codable {                // thats enough for structs to be Codable
    let name: String
    var age: Int
    var mood: Happiness                 // as long as all members are Codable
}

let me = Person(name: "Frank-Peter", age: 60, mood: Happiness.Happy(ofGrade: 5))
let data2 = try PropertyListEncoder().encode(me)
let me2 = try PropertyListDecoder().decode(Person.self, from: data2)
Swift.print("\(me2)")
